@@ -1,7 +1,7 @@
 /*
  * The PEP 484 type hints generator for SIP.
  *
- * Copyright (c) 2016 Riverbank Computing Limited <info@riverbankcomputing.com>
+ * Copyright (c) 2017 Riverbank Computing Limited <info@riverbankcomputing.com>
  *
  * This file is part of SIP.
  *
@@ -43,6 +43,9 @@ static void pyiCtor(sipSpec *pt, moduleDef *mod, classDef *cd, ctorDef *ct,
 static void pyiCallable(sipSpec *pt, moduleDef *mod, memberDef *md,
         overDef *overloads, int is_method, ifaceFileList *defined, int indent,
         FILE *fp);
+static void pyiProperty(sipSpec *pt, moduleDef *mod, propertyDef *pd,
+        int is_setter, memberDef *md, overDef *overloads,
+        ifaceFileList *defined, int indent, FILE *fp);
 static void pyiOverload(sipSpec *pt, moduleDef *mod, overDef *od,
         int overloaded, int is_method, int sec, ifaceFileList *defined,
         int indent, int pep484, FILE *fp);
@@ -54,10 +57,8 @@ static int pyiArgument(sipSpec *pt, moduleDef *mod, argDef *ad, int arg_nr,
         ifaceFileList *defined, KwArgs kwargs, int pep484, FILE *fp);
 static void pyiType(sipSpec *pt, moduleDef *mod, argDef *ad, int out, int sec,
         ifaceFileList *defined, int pep484, FILE *fp);
-static void pyiTypeHint(sipSpec *pt, typeHintDef *thd, moduleDef *mod, int out,
-        ifaceFileList *defined, int pep484, FILE *fp);
 static void pyiTypeHintNode(typeHintNodeDef *node, moduleDef *mod,
-        ifaceFileList *defined, int pep484, FILE *fp);
+        ifaceFileList *defined, int pep484, int rest, FILE *fp);
 static void prIndent(int indent, FILE *fp);
 static int separate(int first, int indent, FILE *fp);
 static void prClassRef(classDef *cd, moduleDef *mod, ifaceFileList *defined,
@@ -284,6 +285,7 @@ static void pyiClass(sipSpec *pt, moduleDef *mod, classDef *cd,
     classDef *nested;
     ctorDef *ct;
     memberDef *md;
+    propertyDef *pd;
 
     separate(TRUE, indent, fp);
     prIndent(indent, fp);
@@ -465,6 +467,27 @@ static void pyiClass(sipSpec *pt, moduleDef *mod, classDef *cd,
         pyiCallable(pt, mod, md, cd->overs, TRUE, *defined, indent, fp);
     }
 
+    for (pd = cd->properties; pd != NULL; pd = pd->next)
+    {
+        first = separate(first, indent, fp);
+
+        if (pd->get != NULL)
+        {
+            if ((md = findMethod(cd, pd->get)) != NULL)
+            {
+                pyiProperty(pt, mod, pd, FALSE, md, cd->overs, *defined,
+                        indent, fp);
+
+                if (pd->set != NULL)
+                {
+                    if ((md = findMethod(cd, pd->set)) != NULL)
+                        pyiProperty(pt, mod, pd, TRUE, md, cd->overs, *defined,
+                                indent, fp);
+                }
+            }
+        }
+    }
+
     /*
      * Keep track of what has been defined so that forward references are no
      * longer required.
@@ -721,6 +744,48 @@ static void pyiCallable(sipSpec *pt, moduleDef *mod, memberDef *md,
 }
 
 
+/*
+ * Generate the type hints for a property.
+ */
+static void pyiProperty(sipSpec *pt, moduleDef *mod, propertyDef *pd,
+        int is_setter, memberDef *md, overDef *overloads,
+        ifaceFileList *defined, int indent, FILE *fp)
+{
+    overDef *od;
+
+    /* Handle each overload. */
+    for (od = overloads; od != NULL; od = od->next)
+    {
+        if (isPrivate(od))
+            continue;
+
+        if (od->common != md)
+            continue;
+
+        if (od->no_typehint)
+            continue;
+
+        prIndent(indent, fp);
+
+        if (is_setter)
+            fprintf(fp, "@%s.setter\n", pd->name->text);
+        else
+            fprintf(fp, "@property\n");
+
+        prIndent(indent, fp);
+
+        fprintf(fp, "def %s", pd->name->text);
+
+        pyiPythonSignature(pt, mod, &od->pysig, TRUE, FALSE, defined,
+                od->kwargs, TRUE, fp);
+
+        fprintf(fp, ": ...\n");
+
+        break;
+    }
+}
+
+
 /*
  * Generate the docstring for a single API overload.
  */
@@ -889,7 +954,7 @@ static void pyiType(sipSpec *pt, moduleDef *mod, argDef *ad, int out, int sec,
 
     if (thd != NULL)
     {
-        pyiTypeHint(pt, thd, mod, out, defined, pep484, fp);
+        pyiTypeHint(pt, thd, mod, out, defined, pep484, FALSE, fp);
         return;
     }
 
@@ -1065,7 +1130,7 @@ static void pyiType(sipSpec *pt, moduleDef *mod, argDef *ad, int out, int sec,
  */
 void prScopedPythonName(FILE *fp, classDef *scope, const char *pyname)
 {
-    if (scope != NULL)
+    if (scope != NULL && !isHiddenNamespace(scope))
     {
         prScopedPythonName(fp, scope->ecd, NULL);
         fprintf(fp, "%s.", scope->pyname->text);
@@ -1348,13 +1413,13 @@ typeHintDef *newTypeHint(char *raw_hint)
 /*
  * Generate a type hint from a /TypeHint/ annotation.
  */
-static void pyiTypeHint(sipSpec *pt, typeHintDef *thd, moduleDef *mod, int out,
-        ifaceFileList *defined, int pep484, FILE *fp)
+void pyiTypeHint(sipSpec *pt, typeHintDef *thd, moduleDef *mod, int out,
+        ifaceFileList *defined, int pep484, int rest, FILE *fp)
 {
     parseTypeHint(pt, thd, out);
 
     if (thd->root != NULL)
-        pyiTypeHintNode(thd->root, mod, defined, pep484, fp);
+        pyiTypeHintNode(thd->root, mod, defined, pep484, rest, fp);
     else
         maybeAnyObject(thd->raw_hint, pep484, fp);
 }
@@ -1364,7 +1429,7 @@ static void pyiTypeHint(sipSpec *pt, typeHintDef *thd, moduleDef *mod, int out,
  * Generate a single node of a type hint.
  */
 static void pyiTypeHintNode(typeHintNodeDef *node, moduleDef *mod,
-        ifaceFileList *defined, int pep484, FILE *fp)
+        ifaceFileList *defined, int pep484, int rest, FILE *fp)
 {
     switch (node->type)
     {
@@ -1385,7 +1450,7 @@ static void pyiTypeHintNode(typeHintNodeDef *node, moduleDef *mod,
 
                 need_comma = TRUE;
 
-                pyiTypeHintNode(thnd, mod, defined, pep484, fp);
+                pyiTypeHintNode(thnd, mod, defined, pep484, rest, fp);
             }
 
             fprintf(fp, "]");
@@ -1394,11 +1459,19 @@ static void pyiTypeHintNode(typeHintNodeDef *node, moduleDef *mod,
         break;
 
     case class_node:
-        prClassRef(node->u.cd, mod, defined, pep484, fp);
+        if (rest)
+            restPyClass(node->u.cd, TRUE, fp);
+        else
+            prClassRef(node->u.cd, mod, defined, pep484, fp);
+
         break;
 
     case enum_node:
-        prEnumRef(node->u.ed, mod, defined, pep484, fp);
+        if (rest)
+            restPyEnum(node->u.ed, TRUE, fp);
+        else
+            prEnumRef(node->u.ed, mod, defined, pep484, fp);
+
         break;
 
     case brackets_node:
@@ -1635,7 +1708,7 @@ static const char *typingModule(const char *name)
 
 
 /*
- * Flatten an unions in a list of nodes.
+ * Flatten any unions in a list of nodes.
  */
 static typeHintNodeDef *flatten_unions(typeHintNodeDef *nodes)
 {
